# Published by k.inaba
#   under NYSL 0.9982 http://www.kmonos.net/nysl/

##################################################
# Parser
##################################################

def parse_app(code)
	def App(n, m)
		return [:app, n, m]
	end

	code =~ /(W+)(w+)/
	return App($1.size, $2.size), $'
end

def parse_abs(code)
	def Abs(n, body)
		n.times { body = [[:abs1, body]] }
		return body[0]
	end

	code =~ /(w+)/
	n = $1.size
	code = $'
	body = []
	while code[0,1] == "W"
		a, code = parse_app(code)
		body << a
	end
	return Abs(n, body), code
end

def parse_prog(code)
	code = code.scan(/[wWv]/) * "" # support ASCII only
	body = []
	while code.size>0
		case code[0,1]
			when "W"
				a, code = parse_app(code)
				body << a
			when "w"
				a, code = parse_abs(code)
				body << a
			when "v"
				code = code[1..-1]
		end
	end
	return body
end

##################################################
# CPS Conversion
##################################################

def cps(body)
	def cps_impl(body, shift=[0,0,0,0])
		# [kont], [arg?], out, succ, w, in
		kont = 1
		body.each do |cmd|
			if cmd[0]==:abs1
				cps_impl(cmd[1], ([0]+shift).map{|s|s+1})
			elsif cmd[0]==:app
				cmd[1] += shift[cmd[1]-1]
				cmd[2] += shift[cmd[2]-1]
			end
			shift.unshift 0
			kont += 1
		end
		body.push [:tailapp, kont, (kont==1 ? 2 : 1)]
	end

	cps_impl(body)
	body = [ [:abs1, [[:in],   [:tailapp,2,1]]],  #in
	         [:putw], #w
	         [:abs1, [[:succ], [:tailapp,2,1]]], #succ
	         [:abs1, [[:out],  [:tailapp,1,2]]], #out
	] + body[0..-2] + [
	    [:app,1,1],
	    [:exit]
	]
	return body
end

##################################################
# Labeling
##################################################

def labeling(body)
	def labeling_impl(body, code_array)
		body.each do |cmd|
			addr = code_array.size
			if cmd[0] == :abs1
				code_array << nil
				bodyAddr = code_array.size
				labeling_impl(cmd[1], code_array)
				nextAddr = code_array.size
				code_array[addr] = [[:abs1, bodyAddr], nextAddr]
			else
				nextAddr = code_array.size+1
				code_array << [cmd, nextAddr]
			end
		end
	end
	code_array = []
	labeling_impl(body, code_array)
	return code_array
end

########################################################
# Evaluator
########################################################

FTrue  = [[[:abs1,[                   # \x->\y->
           [:abs1, [[:tailapp,1,2]]], #   push 'id'
           [:app, 1, 5],              #   id x
           [:tailapp, 3, 1]           #   return
         ]],[:tailapp,2,1]],[]]
FFalse = [[[:abs1,[                   # \x->
           [:tailapp,1,2]             #   id
         ]],[:tailapp,2,1]],[]]

def jmp_eval(code_array)
	i = 0
	e = []
	loop do
		cmd, nxt = code_array[i]
		case cmd[0]
			when :out
				putc e[1].chr
				i = nxt
			when :in
				val = STDIN.getc || e[1]
				i = nxt
				e = [val]+e
			when :succ
				arg = e[1]
				i = nxt
				e = [(arg+1)%256]+e
			when :abs1
				bodyAddr = cmd[1]
				i = nxt
				e = [[bodyAddr,e]]+e
			when :exit
				exit
			when :putw
				i = nxt
				e = [119]+e
			when :app
				_, m, n = cmd
				arg = e[n-1]
				case e[m-1]
					when Integer
						num = e[m-1]
						res = (arg==num ? FTrue : FFalse)
						i = nxt
						e = [res]+e
					else
						cm, em = e[m-1]
						i = cm
						e = [[nxt,e], arg]+em
				end
			when :tailapp
				_, m, n = cmd
				arg = e[n-1]
				cm, em = e[m-1]
				i = cm
				e = [arg]+em
		end
	end
end

########################################################################
# To BF Compiler
########################################################################

N,C_ADR,E_ADR,M,C,I,G,D,T1,T2,T3,T4,T5,T6 = 0,1,2,3,4,5,6,7,8,9,10,11,12,13
VAL = E_ADR
ATOMSIZE = 14
FLAGSIZE = 3

#atom = NEXT C_ADR E_ADR  ++  M C I G D 0 0 0 0
#     | NEXT     0   INT   |  0 0 0 0 0 0 0 0 0
#flag = 0 | 1
#memory = atom* flag*

def move(go, bk, s)
	"[-#{go}+#{bk}]>"*s + "<"*s + go
end
def move_left(d, s)
	move("<"*d, ">"*d, s)
end
def move_right(d, s)
	move(">"*d, "<"*d, s)
end

def goto_G_from_0
	# _ _ _ M C I G * * *
	#   while(G) { G-=1; I+=1; move_left; }
	">>>>>>[-<+<< #{move_left(ATOMSIZE, ATOMSIZE-3)} >>>]<<<<<<"
end

def goback_to_0
	# _ _ _ M C I G * * *
	#   while(I) { I-=1; move_right; }
	">>>>>[-<< #{move_right(ATOMSIZE, ATOMSIZE-3)} >>]<<<<<"
end

def goto_G_from_I
	goback_to_0 + goto_G_from_0
end

def copy(from, to, tmp)
	# precond : from<to<tmp;  cell[to]==cell[tmp]==0
	"#{">"*from}[- #{">"*(to-from)} + #{">"*(tmp-to)} + #{"<"*(tmp-from)}]" +
	"#{">"*(tmp-from)}[- #{"<"*(tmp-from)} + #{">"*(tmp-from)}]" +
	"#{"<"*tmp}"
	# postcond: cell[tmp]=0
end

def copy2(from, to, tmp)
	# precond : to<from<tmp;  cell[to]==cell[tmp]==0
	"#{">"*from}[- #{"<"*(from-to)} + #{">"*(tmp-to)} + #{"<"*(tmp-from)}]" +
	"#{">"*(tmp-from)}[- #{"<"*(tmp-from)} + #{">"*(tmp-from)}]" +
	"#{"<"*tmp}"
	# postcond: cell[tmp]=0
end

def goto_next
	# N _ _ M C I G D * *
	#   G=N; goto_G_from_I;
	copy(N,G,T1) + goto_G_from_I
end
def deref_D
	# N _ _ M C I G D * *
	#   while(D) { D-=1; goto_next; }
	">>>>>>>[- <<<<<<< #{goto_next} >>>>>>>]<<<<<<<"
end


def to_bf(c)
	puts "+#{">"*FLAGSIZE}+#{"<"*FLAGSIZE}["
	c.each_with_index do |cmd,i|
		puts "  #{">"*FLAGSIZE}[-"
		cmd, nxt = cmd
		case cmd[0]
			when :in
				# TODO: EOF check
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_0
				puts "     >>>>[-<<<<+>>>>]<<,<<" # _ _ _ = C 0 <in>,  C=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				puts "     #{">"*((i+1)*FLAGSIZE+ATOMSIZE)}" # comeback
				if i<nxt then puts "     #{">"*(nxt-i)*FLAGSIZE}+#{"<"*(nxt-i)*FLAGSIZE}"
				         else puts "     #{"<"*(i-nxt)*FLAGSIZE}+#{">"*(i-nxt)*FLAGSIZE}" end
			when :out
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_0
				puts "     #{copy(N,G,T1)}" # G := N
				puts "     " + goto_G_from_I
				puts "     >>.<<" # putchar
				puts "     " + goback_to_0
				puts "     #{">"*((i+1)*FLAGSIZE+ATOMSIZE)}" # comeback
				if i<nxt then puts "     #{">"*(nxt-i)*FLAGSIZE}+#{"<"*(nxt-i)*FLAGSIZE}"
				         else puts "     #{"<"*(i-nxt)*FLAGSIZE}+#{">"*(i-nxt)*FLAGSIZE}" end
			when :putw
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_0
				puts "     >>>>[-<<<<+>>>>]<<#{"+"*119}<<" # _ _ _ = C 0 119,  C=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				puts "     #{">"*((i+1)*FLAGSIZE+ATOMSIZE)}" # comeback
				if i<nxt then puts "     #{">"*(nxt-i)*FLAGSIZE}+#{"<"*(nxt-i)*FLAGSIZE}"
				         else puts "     #{"<"*(i-nxt)*FLAGSIZE}+#{">"*(i-nxt)*FLAGSIZE}" end
			when :abs1
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_I
				puts "     >>>>[-<<+<<+>>>>]<<<#{"+"*cmd[1]}<" # _ _ _ = C cmd[1] C,  C=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				puts "     #{">"*((i+1)*FLAGSIZE+ATOMSIZE)}" # comeback
				if i<nxt then puts "     #{">"*(nxt-i)*FLAGSIZE}+#{"<"*(nxt-i)*FLAGSIZE}"
				         else puts "     #{"<"*(i-nxt)*FLAGSIZE}+#{">"*(i-nxt)*FLAGSIZE}" end
			when :exit
				puts "     #{"<"*(i+1)*FLAGSIZE}-#{">"*(i+1)*FLAGSIZE}"
			when :succ
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_0
				puts "     " + goto_next
				puts "     #{copy(VAL,T3,T4)}"  # T3 := VAL
				puts "     #{">"*T3}+#{"<"*T3}" # T3 ++
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_I
				puts "     >>>>[-<<<<+>>>>] #{">"*(T3-C)}[-#{"<"*(T3-VAL)}+#{">"*(T3-VAL)}]#{"<"*T3}" # _ _ _ = C 0 T3,  C=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				puts "     #{">"*((i+1)*FLAGSIZE+ATOMSIZE)}" # comeback
				if i<nxt then puts "     #{">"*(nxt-i)*FLAGSIZE}+#{"<"*(nxt-i)*FLAGSIZE}"
				         else puts "     #{"<"*(i-nxt)*FLAGSIZE}+#{">"*(i-nxt)*FLAGSIZE}" end
			when :app
				# ToDo: Integer App
				_, m, n = cmd; m-=1; n-=1
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_0
				puts "     #{">"*D}#{"+"*m}#{"<"*D}"
				puts "     " + deref_D # deref m
				puts "     #{copy(C_ADR, T2, T6)}" # T2 := e[m].code
				puts "     #{copy(E_ADR, T3, T6)}" # T3 := e[m].env
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_I
				puts "     #{">"*D}#{"+"*n}#{"<"*D}"
				puts "     " + deref_D # deref n
				puts "     #{copy(C_ADR, T4, T6)}" # T4 := e[n].code
				puts "     #{copy(E_ADR, T5, T6)}" # T5 := e[n].env
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_I
				puts "     #{">"*T3}[-#{"<"*T3}+#{">"*T3}]"  # N<-T3
				puts "     >[-#{"<"*T3}+#{">"*T3}]"  # C_ADR<-T4
				puts "     >[-#{"<"*T3}+#{">"*T3}]#{"<"*T5}"  # E_ADR<-T5
				puts "     >>>#{move_left(ATOMSIZE, ATOMSIZE-3)}>>+<<<<<"
				puts "     #{copy2(M,N,T1)}"
				puts "     >#{"+"*nxt}<"
				puts "     #{copy2(C,E_ADR,T1)}" #  put [M,  #{nxt}, e]
				puts "     >>>+<<<" # M+=1
				puts "     >>>>[-]<<<<" # C:=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				# nxt = T2
				# FLG T2 I
 				puts "     #{">"*T2}[-#{">"*(ATOMSIZE-T2)}>>>>+<<<<#{"<"*(ATOMSIZE-T2)}]#{">"*(ATOMSIZE-T2)}>>>>"
				# while(T2) {go_next; T2-- I++}
				puts "     [- [->>>+<<<]>+[->>>+<<<]>> ] <+>>"
				# while(I) {go_prev; --I}
				puts "     [- [-<<<+>>>] <<<] <<<<<"
				puts "     #{">"*((i+1)*FLAGSIZE)}" # comeback
			when :tailapp
				_, m, n = cmd; m-=1; n-=1
				puts "     #{"<"*((i+1)*FLAGSIZE+ATOMSIZE)}" # focus 0
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_0
				puts "     #{">"*D}#{"+"*m}#{"<"*D}"
				puts "     " + deref_D # deref m
				puts "     #{copy(C_ADR, T2, T6)}" # T2 := e[m].code
				puts "     #{copy(E_ADR, T3, T6)}" # T3 := e[m].env
				puts "     #{copy(C,G,T1)}" # G := C
				puts "     " + goto_G_from_I
				puts "     #{">"*D}#{"+"*n}#{"<"*D}"
				puts "     " + deref_D # deref n
				puts "     #{copy(C_ADR, T4, T6)}" # T4 := e[n].code
				puts "     #{copy(E_ADR, T5, T6)}" # T5 := e[n].env
				puts "     >>>+<<<" # M+=1
				puts "     #{copy(M,G,T1)}" # G := M
				puts "     " + goto_G_from_I
				puts "     #{">"*T3}[-#{"<"*T3}+#{">"*T3}]"  # N<-T3
				puts "     >[-#{"<"*T3}+#{">"*T3}]"  # C_ADR<-T4
				puts "     >[-#{"<"*T3}+#{">"*T3}]#{"<"*T5}"  # E_ADR<-T5
				puts "     >>>>[-]<<<<" # C:=0
				puts "     #{copy(M,C,T1)}" # C := M
				puts "     " + goback_to_0
				# nxt = T2
				# FLG T2 I
 				puts "     #{">"*T2}[-#{">"*(ATOMSIZE-T2)}>>>>+<<<<#{"<"*(ATOMSIZE-T2)}]#{">"*(ATOMSIZE-T2)}>>>>"
				# while(T2) {go_next; T2-- I++}
				puts "     [- [->>>+<<<]>+[->>>+<<<]>> ] <+>>"
				# while(I) {go_prev; --I}
				puts "     [- [-<<<+>>>] <<<] <<<<<"
				puts "     #{">"*((i+1)*FLAGSIZE)}" # comeback
		end
		puts "  ]"
	end
	puts "#{"<"*FLAGSIZE*c.size}]"
end


c = labeling(cps(parse_prog([*$<]*"")))
to_bf(c)
















__END__

require 'pp'
code = <<EOG
ww
 WWw
v
wwww    \abcd
 WWWww     s=b(c)
 Www       t=s(d)
 WWWWWWwwww u=a(c)
 Www  z=u(t)
v
w
 WWwww      Z(id)
 Wwwww       (id)
 Wwwwwww     (out)
 Wwwwwwwwww  (119)
EOG
code = <<EOG
WWWWw
WWWWWw
WWWWWWw
WWWWwww
WWWWWwww
WWWWWWwww
EOG
code = <<EOG
WWwww    : x
WWWWwwww : false
Wwwwww : GRASS_LETTER
Wwww   : x
vwWWWWWWww : out
EOG
code = <<EOG
ɑAƂ܂wWWwwwwWWww
EOG
code = <<EOG
xRrl[^wwWWwwWwwvwwWWWwWWWwvwWWwWwv
EOG
code = <<EOG
wwwwvwvwwWWwvwwWwwvwwwwWWWwwWwwWWWWWWwwwwWw
wvwWWwWwwvwWWWwwwwwWwwwwwwWWwWWWwWWWWWWwW
WWWWWWWwwwWwwWWWWWWWWWWwWwwwwwWWWWWWWW
WWWwwwwwWWWWWWWWWWWWwwwwWWWWWWWWWWWWW
wwwWWWWWWWWWWWWWWwwwWWWWWWWWWWWWWWWwW
WWWWWWWWWWWWWWWWWwwwWwwwwwwwwwwwwwwWWWW
WWWWWWWWWWWWWWWwwwwwwwwWwwWWWWWWWWWWW
WWWWWWWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwwWwwww
wwwwwwwwwwwwwwwww@@@@@@@@@@@@@wwwwwwwwWWwwwwwww
wwwwwwwwwwwwwwwww@@@@@@@ @ ́@@ wwwwwWWWWWWWWWW
WWWWWwWwwwWWWW@@@@@@@@@ WWWWWWWwwwWwwWW
WWWWWWWWWWwwww@@@@@@@́@@  wWwwwwwwwWWWWWWW
WWWWWWWWWWwwww@@@@@@@@@@wwwWwwWWWWWWWWW
WWWWWWWWWwwwww@@@@ @@@@@@@wwwwWwwWWWWWWWW
WWWWWWWWWWWWW@@@@@@@@@@@WWWWWWWWWwwwwww
wwwwwWwwWWWWWWW@@@ @@@@@@@WWWWWWWWWwwwwww
wwwwwwwWwwwwwwwww@@@@@@@@@@@@ wwwwwwWWWWWWWWW
WWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
wwwwwwWwwwwwwwwwwwwwwWWwwwwwwwwwWWWwwWWWWwww
wwwwwwwwwwwwWWWWWwwWWWWWWwwwwWWWWWWWwwWWW
WWWWWwwwwWWWWWWWWWwwWWWWWWWWWWwwwwwwwww
wwwwWWWWWWWWWWWWWWWWWWWWWWWWWWWWwwwwww
wwwwwWwwwWWwwwwwwwwwwwwwwwwwwWWWwwWWWWwwwwww
wwwwwwwwwwwwwwwwwwWWWWWwwWWWWWWwwwwwwwWWWW
WWWwwWWWWWWWWwwwwwwWWWWWWWWWwwWWWWWWWW
WWwwwwwwWWWWWWWWWWWwwwwwwwwwwwwwwwwwwwwwww
EOG
code = <<EOG
wWWwwww
EOG
code = <<EOG
wWWwwwwWWWwWWWWw
EOG
code = <<EOG
wWWWwwwwWWWw
EOG
code = <<EOG
ɑAƂ܂wWWwwwwWWww
EOG
code = <<EOG
wv	: id
wwWWwWWWwv	: \ f x -> f (f x)
Wwwww	: + 2
WWw	: + 4
WWWw	: + 8
WWWWw	: + 16
WWWWWw	: + 32
WWWWWWw	: + 64
WWWWWWWw	: + 128
Wwwwwwwwwwwww	: du + 128 
WWWWw	: du + 160
WWWWWWWw	: du + 168
WWWWWWWWWWWWWWw	: du + 169 = SP
WWWWWWWWWWWww	: du + 170 = !
WWWWWWWWWWww	: du + 177
WWWWWWWWWWWWw	: du + 181 = ,
WWWWWWWWWWww	: du + 209 = H
WWWWWWWWWWwwwwww	: du + 232
WWWWWWWWWWWWWWWw	: du + 236
WWWWWWWWWWWWWWWWWWWWWw	: du + 237 = d
WWWWWWWWWWWWWWWWWWww	: du + 238 = e
WWWWWWWWWWWWWWWWWww	: du + 245 = l
WWWWWWWWWWWWWWWWWwwwww	: du + 248 = o
WWWWWWWWWWWWWWWWWWWWww	: du + 249
WWWWWWWWWWWWWWWWWWWWWWw	: du + 251 = r
WWWWWWWWWWWWWWWWWWWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwwww	: id out
Wwwwwwwwwww	: out H
WWwwwwwww	: out e
WWWwwwwwww	: out l
WWWWw	: out l
WWWWWwwwwwwww	: out o
WWWWWWwwwwwwwwwwwwwwww	: out ,
WWWWWWWwwwwwwwwwwwwwwwwwwww	: out SP
WWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww	: out du
WWWWWWWWWwwww	: out o
WWWWWWWWWWwwwwwwwwwww	: out r
WWWWWWWWWWWwwwwwww	: out l
WWWWWWWWWWWWwwwwwwwwwwwwwwwwww	: out d
WWWWWWWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwww	: out !
EOG

to_bf([[[:abs1, 1], 3],
   [[:in], 2],
   [[:tailapp, 2, 1], 3],
 [[:putw], 4],
 [[:abs1, 5], 7],
   [[:succ], 6],
   [[:tailapp, 2, 1], 7],
 [[:abs1, 8], 10],
   [[:out], 9],
   [[:tailapp, 1, 2], 10],
 [[:abs1, 11], 14],
   [[:app, 4, 5], 12],
   [[:app, 4, 1], 13],
   [[:tailapp, 3, 1], 14],
 [[:app, 1, 1], 15],
 [[:exit], 16]]
)
exit

to_bf(
[[[:abs1, 1], 3],
   [[:in], 2],
   [[:tailapp, 2, 1], 3],
 [[:putw], 4],
 [[:abs1, 5], 7],
   [[:succ], 6],
   [[:tailapp, 2, 1], 7],
 [[:abs1, 8], 10],
   [[:out], 9],
   [[:tailapp, 1, 2], 10],
 [[:app, 1, 3], 11],
# [[:app, 1, 1], 12],
 [[:exit], 13]]
)
exit

to_bf([
	[[:abs1, 1], 5],
		[[:putw],2],
		[[:putw],3],
		[[:out],4],
		[[:exit],5],
	[[:tailapp, 1, 1], 1],
	[[:exit],6],
])

